import sys
from libcpp.vector cimport vector
import numpy as np
from scipy.stats import poisson
from numpy import log
from itertools import count

from libc.math cimport log as clog
from libc.math cimport round as cround

try:
    from functools import lru_cache
except ImportError:
    from functools32 import lru_cache


@lru_cache()
def compute_window_score(int i, _poisson):

    cdef float window_score
    cdef float p_value

    if i < _poisson.mean():
        # sys.stderr.write(" ".join([str(e) for e in [ "i of", i, "gives poisson score of", 0, "\n" ]]))
        return 0

    p_value = _poisson.pmf(i)

    if p_value > 0:
        # sys.stderr.write(" ".join([str(e) for e in ["i of", i, "gives poisson score of", -log(p_value), "and p_value of", p_value, "\n"] ]))
        window_score = -log(p_value)
    else:
        # log of zero not defined
        # sys.stderr.write(" ".join([str(e) for e in ["i of", i, "gives poisson score of", -log(p_value), "and p_value of", p_value, "\n"] ]))
        window_score = 1000

    # sys.stderr.write("Returning window score of " + str(window_score) + "\n")
    return window_score




def generate_cumulative_distribution(vector[float] island_expectations):

    cdef:
        int i
        int l = island_expectations.size()
        double [::1] cumulative
        float partial_sum = 0


    cumulative_arr = np.zeros(l, dtype=float)
    cumulative = cumulative_arr

    for i in range(1, l + 1):

        compliment = l - i
        partial_sum += island_expectations[compliment]
        cumulative[compliment] = partial_sum

    partial_sum += island_expectations[l]
    cumulative[1] = partial_sum

    return cumulative


def update_island_expectations(vector[float] island_expectations, int scaled_score, int bin_size, float average_window_readcount, int island_enriched_threshold, float gap_contribution):

    cdef:
        float WINDOW_P_VALUE = 0.20
        float BIN_SIZE = 0.001
        int E_VALUE = 1000
        float E_VALUE_THRESHOLD = E_VALUE * .0000001
        int i = island_enriched_threshold
        float temp
        int index

    _poisson = poisson(average_window_readcount)
    current_max_scaled_score = island_expectations.size() - 1
    if scaled_score > current_max_scaled_score:
        #index is the scaled_score
        for index in range(current_max_scaled_score + 1, scaled_score + 1):
            temp=0.0


            current_island = int(cround(index - compute_window_score(i, _poisson) / BIN_SIZE))

            while (current_island >= 0):
                island_expectation = island_expectations[current_island]

                temp += _poisson.pmf(i) * island_expectation
                i += 1
                if i == 500:
                    break

                current_island = int(cround(index - compute_window_score(i, _poisson) / BIN_SIZE))

                temp *= gap_contribution
            island_expectations.push_back(temp)


    return island_expectations


def compute_score_threshold(average_window_readcount,
                            island_enriched_threshold,
                            gap_contribution, boundary_contribution,
                            genome_length_in_bins, bin_size):
    # type: (float, int, float, float, float) -> float
    """
    What does island_expectations do?
    """

    cdef:
        int current_scaled_score
        int l
        int interval
        float partial_cumu
        float e

        float WINDOW_P_VALUE = 0.20
        float BIN_SIZE = 0.001
        int E_VALUE = 1000
        float E_VALUE_THRESHOLD = E_VALUE * .0000001


    required_p_value = poisson.pmf(island_enriched_threshold,
                                   average_window_readcount)
    # print("required_p_value", required_p_value)
    prob = boundary_contribution * required_p_value
    # print("prob", prob)

    score = -log(required_p_value)
    # print("score", score)

    current_scaled_score = int(round(score / BIN_SIZE))
    # print("current_scaled_score", current_scaled_score)

    cdef vector[float] island_expectations = vector[float](current_scaled_score + 1)

    island_expectations[0] = boundary_contribution * genome_length_in_bins / gap_contribution

    island_expectations[current_scaled_score] = prob * genome_length_in_bins
    # print("len(island_expectations)", island_expectations.size())
    # print(island_expectations)

    current_max_scaled_score = current_scaled_score

    interval = 1000 # 1 / BIN_SIZE)
    partial_cumu = 0.0

    while (partial_cumu > E_VALUE_THRESHOLD or partial_cumu < 1e-100):

        current_scaled_score += interval
        island_expectations = update_island_expectations(island_expectations, current_scaled_score, bin_size, average_window_readcount, island_enriched_threshold, gap_contribution)
        current_expectation = island_expectations[current_scaled_score]

        l = island_expectations.size()
        if l > interval:
            partial_cumu = sum(island_expectations[i] for i in range(l - interval + 1, l))
        else:
            partial_cumu = sum(island_expectations)

    cumulative = generate_cumulative_distribution(island_expectations)
    # print(list(cumulative))

    score_threshold = 0
    for (i, e) in enumerate(cumulative):
        if e < E_VALUE:
            score_threshold = (i - 1) * BIN_SIZE
            break

    return score_threshold



def compute_enriched_threshold(average_window_readcount):
    # type: (float) -> int
    """
    Computes the minimum number of tags required in window for an island to be enriched.
    """

    cdef:

        float WINDOW_P_VALUE = 0.20
        float BIN_SIZE = 0.001
        int E_VALUE = 1000
        float E_VALUE_THRESHOLD = E_VALUE * .0000001

    current_threshold, survival_function = 0, 1
    while True:
        survival_function -= poisson.pmf(current_threshold,
                                         average_window_readcount)
        current_threshold += 1
        if survival_function <= WINDOW_P_VALUE:
            break

    island_enriched_threshold = current_threshold

    return island_enriched_threshold


def compute_gap_factor(island_enriched_threshold,
                       gap_intervals_allowed, poisson_distribution_parameter):
    # type: (int, int, float) -> float

    max_gap_score = 1.0
    gap_factor = single_gap_factor(island_enriched_threshold,
                                   poisson_distribution_parameter)
    max_gap_score += sum([pow(gap_factor, i)
                          for i in range(1, gap_intervals_allowed + 1)])
    return max_gap_score


def single_gap_factor(island_enriched_threshold,
                      poisson_distribution_parameter):
    # type: (int, float) -> float

    poisson_scores = [poisson.pmf(i, poisson_distribution_parameter)
                      for i in range(island_enriched_threshold)]
    return sum(poisson_scores)


def compute_boundary(island_enriched_threshold, gap_intervals_allowed,
                     average):
    # type: (int, int, float) -> float

    single_gap = single_gap_factor(island_enriched_threshold, average)
    single_boundary_score = pow(single_gap, gap_intervals_allowed + 1)
    start_and_end_score = single_boundary_score * single_boundary_score

    return start_and_end_score


def compute_background_probabilities(total_chip_count, bin_size, effective_genome_fraction, gaps_allowed):
    # type: (int, Namespace) -> Tuple[float, int, float]

    print("total_chip_count", total_chip_count)
    print("bin_size", bin_size)
    print("effective_genome_fraction", effective_genome_fraction)

    average_window_readcount = total_chip_count * (bin_size / float(effective_genome_fraction))
    print("average_window_readcount", average_window_readcount)

    island_enriched_threshold = compute_enriched_threshold(average_window_readcount)
    print("island_enriched_threshold", island_enriched_threshold)

    gap_contribution = compute_gap_factor(island_enriched_threshold, gaps_allowed, average_window_readcount)
    print("gap_contribution", gap_contribution)

    boundary_contribution = compute_boundary(island_enriched_threshold, gaps_allowed, average_window_readcount)
    print("boundary_contribution", boundary_contribution)

    genome_length_in_bins = effective_genome_fraction / bin_size
    print("genome_length_in_bins", genome_length_in_bins)

    score_threshold = compute_score_threshold(average_window_readcount, island_enriched_threshold, gap_contribution, boundary_contribution, genome_length_in_bins, bin_size)
    print("score_threshold", score_threshold)

    return score_threshold, island_enriched_threshold, average_window_readcount

# ('total_chip_count', 9999)
# ('bin_size', 200)
# ('effective_genome_fraction', 2476555186.4)
# ('average_window_readcount', 0.0008074926054472355)
# ('island_enriched_threshold', 1)
# ('gap_contribution', 3.9951596055200813)
# ('boundary_contribution', 0.9935608797169523)
# ('genome_length_in_bins', 12382775.932)
# ('required_p_value', 0.0008068408243290163)
# ('prob', 0.0008016454792118883)
# ('score', 7.122384152855723)
# current_scaled_score
# ('len(island_expectations)', 7124)
# (7122, 9926.5966796875)
# (7123, 3079487.0)



import pandas as pd
import pyranges as pr
from epic2.src.genome_info import sniff
from epic2.src.reads_to_bins import files_to_bin_counts

# note that


def file_to_unbinned_ranges(f, args, _):

    file_format = sniff(f)

    names = "Chromosome Start End Strand".split()
    if file_format == "bed" or "bed.gz":
        df = pd.read_csv(f, sep="\t", header=None, usecols=[0, 1, 2, 5], names=names)
        gr = pr.PyRanges(df)
    elif file_format == "bedpe":
        df = pd.read_csv(f, sep="\t", header=None, usecols=[0, 1, 5, 9], names=names)
        gr = pr.PyRanges(df)
    else:
        gr = pr.read_bam(f)

    if args["drop_duplicates"]:
        gr = gr.drop_duplicate_positions()

    return gr



def file_to_binned_ranges(f, args, datatype):

    chromosome_data = {}
    # counts = 0
    bin_counts, _ = files_to_bin_counts([f], args, datatype)

    for c, v in bin_counts.items():
        starts, scores = v
        ends = starts + args["bin_size"]
        df = pd.DataFrame({"Start": starts, "End": ends, "Score": scores})
        df.insert(0, "Chromosome", c)
        chromosome_data[c] = df
        # counts += len(df)

    return pr.PyRanges(chromosome_data)





def files_to_coverage(files, args):

    # have same writing-mechanism for both
    # just need to make bins into pyranges before
    # bigwig
    if args["raw"]:
        file_to_ranges = file_to_unbinned_ranges
    else:
        file_to_ranges = file_to_binned_ranges


    # read data and make ranges of coverage
    file_ranges = {}
    for f in files:
        ranges = file_to_ranges(f, args, "ChIP")
        file_ranges[f] = ranges


    # create coverage
    file_coverage = {}
    for n, gr in file_ranges.items():
        file_coverage[n] = gr

    return file_coverage


def main(args):

    # TODO: need to create coverage of file if raw
    # else cluster Scores of binned

    treatment_ranges = files_to_coverage(args["treatment"], args)

    if args.get("control"):
        control_ranges = files_to_coverage(args["control"], args)
        control_sum = pr.concat(control_ranges.values())

    treatment_sum = pr.concat(treatment_ranges.values())

    print("treatment_sum", treatment_sum)



from sys import argv
import sys
import os

from epic2.src.reads_to_bins import files_to_bin_counts
from epic2.src.find_islands import find_islands, compute_fdr, write_islands, add_chip_count_to_islands
from epic2.src.genome_info import egl_and_chromsizes

from collections import OrderedDict


def _main(args):

    args["drop_duplicates"] = int(not args["keep_duplicates"])

    import logging

    if args["quiet"]:
        level = logging.CRITICAL
    else:
        level = logging.INFO

    logging.basicConfig(
        level=level,
        format='%(message)s ',
        datefmt='%a, %d %b %Y %H:%M:%S',
        stream=sys.stderr)

    output = args["output"]
    if output:
        outfile_directory = os.path.dirname(output)
        if outfile_directory and not os.path.exists(outfile_directory):
            os.makedirs(outfile_directory)

    effective_genome_length, chromsizes = egl_and_chromsizes(args)
    args["chromsizes_"] = chromsizes
    args["effective_genome_size"] = effective_genome_length

    c_bins_counts, chip_count_before = files_to_bin_counts(
        args["treatment"], args, "ChIP")
    chip_count = sum(sum(counts) for _, counts in c_bins_counts.values())

    logging.info(
        "\nValid ChIP reads: {} ({} before out of bounds removal)\n".format(
            chip_count, chip_count_before))

    if args["original_statistics"]:
        from epic2.src.SICER_stats import compute_score_threshold
    else:
        from epic2.src.SICER_stats2 import compute_score_threshold

    score_threshold, island_enriched_threshold, average_window_readcount = compute_score_threshold(
        chip_count, args["bin_size"], effective_genome_length,
        args["gaps_allowed"] * args["bin_size"], args["e_value"])

    logging.info(
        "Number of tags in a window: {}\n".format(island_enriched_threshold))

    islands = find_islands(c_bins_counts, args["gaps_allowed"],
                           args["bin_size"], score_threshold,
                           island_enriched_threshold, average_window_readcount)
    new_islands = add_chip_count_to_islands(islands, c_bins_counts, args)

    logging.info("Number of islands found: {}\n".format(
        sum(len(i) for i in islands.values())))

    if args["control"]:

        b_bins_counts, background_count_before = files_to_bin_counts(
            args["control"], args, "Input")
        background_count = sum(
            sum(counts) for _, counts in b_bins_counts.values())
        logging.info(
            "\nValid Background reads: {} ({} before out of bounds removal)\n".
            format(background_count, background_count_before))

        if args["original_algorithm"]:
            cc, bc = chip_count_before, background_count_before
        else:
            cc, bc = chip_count, background_count

        compute_fdr(new_islands, b_bins_counts, cc, bc,
                    effective_genome_length,
                    args["false_discovery_rate_cutoff"], args["output"])
    else:
        write_islands(new_islands, average_window_readcount,
                      args["false_discovery_rate_cutoff"], args["e_value"],
                      args["output"])

        if args["original_algorithm"]:
            cc = chip_count_before
        else:
            cc = chip_count

    return c_bins_counts



import sys
import logging
from math import pi, log, exp, ceil, fabs


class Background_island_probscore_statistics:
    #  External genomeLength and gapSize are in units of bps
    #  Internal genomeLength and gapSize are in units of windows
    #  only look at enrichment!
    def __init__(self, total_tags, windowSize, gapSize, window_pvalue,
                 genomeLength, bin_size):
        self.tag_density = total_tags * 1.0 / genomeLength
        self.window_size = windowSize
        # In bps.
        assert (gapSize % windowSize == 0)
        # gap is in bps
        self.gap_size = gapSize / windowSize
        # maximum number of windows allowed in a gap
        self.genome_length = int(ceil(float(genomeLength) / windowSize))
        self.average = self.tag_density * windowSize
        self.bin_size = bin_size

        # Precalculate the poisson, cumulative poisson values up to max (500, 2*self.average) .
        self.max_index = max(500, int(2 * self.average))
        #print self.average, self.max_index;
        #self.fact=[];
        self.poisson_value = []
        self.window_score = []
        self.window_scaled_score = []
        for index in xrange(self.max_index):
            # self.fact.append(self.factorial(index));
            prob = self.poisson(index, self.average)
            self.poisson_value.append(prob)
            if (index < self.average):  # only want to look at enrichment
                self.window_score.append(0)
                self.window_scaled_score.append(0)
            else:
                if prob > 0:
                    self.window_score.append(-log(prob))
                    #scaled_score =int(-log(prob)/self.bin_size);
                    scaled_score = int(round(-log(prob) / self.bin_size))
                    self.window_scaled_score.append(scaled_score)
                else:  #prob is too small and deemed 0 by the system
                    self.window_score.append(1000)
                    scaled_score = int(round(1000 / self.bin_size))
                    self.window_scaled_score.append(scaled_score)
            #print index, self.poisson_value[index], self.window_score[index];
        self.max_index = len(self.poisson_value)
        #print "max_index ", self.max_index;
        # gap_contribution needs min_tags_in_window
        # So the position of this line is critical.
        self.min_tags_in_window = 0
        sf = 1

        # print("self.average=", self.average)
        # print("self.poisson_value[0]=", self.poisson_value[0])
        # print("window_pvalue", window_pvalue)
        #print "self.poisson_value[0]=", self.poisson_value[0];
        while (sf > window_pvalue):
            #print self.min_tags_in_window, sf;
            # print("poisson", self.min_tags_in_window, self.poisson_value[self.min_tags_in_window])
            sf -= self.poisson_value[self.min_tags_in_window]
            self.min_tags_in_window += 1
        #An alternative approach that uses the scipy package,
        #poisson.sf (n, lambda) = \sum_{i= n+1}^{\infty} p(i, lambda)
        #self.min_tags_in_window = int(self.average);
        #while (scipy.stats.poisson.sf(self.min_tags_in_window-1) > window_pvalue):
        #	self.min_tags_in_window += 1;

        #print "Window read count threshold: ", self.min_tags_in_window;

        self.gap_contribution = self.gap_factor()
        self.boundary_contribution = self.boundary()

        # print("self.gap_contribution", self.gap_contribution)
        # print("self.boundary_contribution", self.boundary_contribution)

        self.cumulative = []
        # new method, first fill the lowest score.
        prob = self.boundary_contribution * self.poisson_value[
            self.min_tags_in_window]
        score = -log(self.poisson_value[self.min_tags_in_window])
        #scaled_score = int(score/self.bin_size);
        scaled_score = int(round(score / self.bin_size))
        self.island_expectation = [0] * (scaled_score + 1)

        # print("island_expectation length", (scaled_score+1))
        self.island_expectation[scaled_score] = prob * self.genome_length
        # print("self.island_expectation[scaled_score]", self.island_expectation[scaled_score])
        # print("scaled_score", scaled_score)

        self.island_expectation[
            0] = self.boundary_contribution * self.genome_length / self.gap_contribution

        self.root = self.find_asymptotics_exponent()
        #print "Exponent for Asymptotics: ", self.root;

    def factorial(self, m):
        value = 1.0
        if m != 0:
            while m != 1:
                value = value * m
                m = m - 1
        return value

    # Return the log of a factorial, using Srinivasa Ramanujan's approximation
    def factln(self, m):
        if m < 20:
            value = 1.0
            if m != 0:
                while m != 1:
                    value = value * m
                    m = m - 1
            return log(value)
        else:
            return m * log(m) - m + log(m * (1 + 4 * m *
                                             (1 + 2 * m))) / 6.0 + log(pi) / 2

    def poisson(self, i, average):
        if i < 20:
            return exp(-average) * average**i / self.factorial(i)
        else:
            exponent = -average + i * log(average) - self.factln(i)
            return exp(exponent)

    """
		gap is in the unit of windows. In each window in the gap, the
		window could have 0, 1, min_tags_in_windows-1 tags.
		say gap = 1, min_tags_in_window= 2, gap_factor = 1 +
		poission(0,a) + poisson(1, a), where 1 represents no gap,
		poisson(0,a) represents a window with 0 tag,
		poisson(1,a) represents a window with 1 tag,
		The gap contribution from each window is not independent
	"""

    def single_gap_factor(self):
        my_gap_factor = 0
        for i in xrange(self.min_tags_in_window):
            my_gap_factor += self.poisson_value[i]
        return my_gap_factor

    # gap contribution is bigger than 1
    def gap_factor(self):
        if self.gap_size == 0:
            return 1
        else:
            i = 1
            gap_contribution = 1
            # contribution from no gap
            my_gap_factor = self.single_gap_factor()
            # print("self.gap_size " * 100, self.gap_size)
            for i in range(1, int(self.gap_size + 1)):
                gap_contribution += pow(my_gap_factor, i)
            return gap_contribution

    def boundary(self):
        """
		The condition for boundary is a continuous region of
		unqualified windows longer than gap
		"""
        temp = self.single_gap_factor()
        temp = pow(temp, self.gap_size + 1)
        return temp * temp
        # start & end

    #forward method that memorize the calculated results.
    def background_island_expectation(self, scaled_score):
        current_max_scaled_score = len(self.island_expectation) - 1
        if scaled_score > current_max_scaled_score:
            #index is the scaled_score
            for index in range(current_max_scaled_score + 1, scaled_score + 1):
                temp = 0.0
                #i is the number of tags in the added window
                i = self.min_tags_in_window
                while (int(
                        round(index - self.window_score[i] / self.bin_size)) >=
                       0):
                    #while ( (index - self.window_scaled_score[i])>=0):
                    temp += self.poisson_value[i] * self.island_expectation[int(
                        round(index - self.window_score[i] / self.bin_size))]
                    #temp += self.poisson_value[i]* self.island_expectation[index - self.window_scaled_score[i]];
                    i += 1
                temp *= self.gap_contribution
                self.island_expectation.append(temp)
                #print index, temp, self.island_expectation[index];
        return self.island_expectation[scaled_score]

    def generate_cumulative_dist(self, outfile=""):
        """
		Generate cumulative distribution: a list of tuples (bins, hist).
		"""
        self.cumulative = [0] * len(self.island_expectation)
        partial_sum = 0.0
        for index in range(1, len(self.island_expectation) + 1):
            complimentary = len(self.island_expectation) - index
            partial_sum += self.island_expectation[complimentary]
            # The end is outside of the index
            self.cumulative[complimentary] = partial_sum

        if outfile != "":
            fixpoint = int(len(self.island_expectation) / 2)
            outf = open(outfile, "w")
            outline = "# Score" + "\t" + "Expect # islands" + "\t" + "Cumulative # Islands" + "\t" + "Asymptotics" + "\n"
            outf.write(outline)
            for index in xrange(len(self.island_expectation)):
                outline = str(index * self.bin_size) + "\t" + str(
                    self.island_expectation[index]) + "\t" + str(
                        self.cumulative[index]) + "\n"
                #outline = str(index * self.bin_size) + "\t" + str(self.island_expectation[index])+ "\t" +str(self.cumulative[index]) + "\t" + str(self.cumulative[fixpoint] * exp(-self.root*(self.cumulative[index]-self.cumulative[fixpoint]))) + "\n";
                outf.write(outline)
            outf.close()

    def find_island_threshold(self, e_value_threshold):
        """
		average is the average number of tags in a window:
		opt.tag_density * opt.window_size
		This one allows single-window islands.
		Returns the island threshold
		"""
        threshold = .0000001 * e_value_threshold
        current_scaled_score = len(self.island_expectation) - 1
        current_expectation = self.island_expectation[-1]
        assert (current_expectation ==
                self.island_expectation[current_scaled_score])
        interval = int(1 / self.bin_size)
        if len(self.island_expectation) > interval:
            partial_cumu = sum(self.island_expectation[-interval:-1])
        else:
            partial_cumu = sum(self.island_expectation)
        while (partial_cumu > threshold or partial_cumu < 1e-100):
            current_scaled_score += interval
            current_expectation = self.background_island_expectation(
                current_scaled_score)
            if len(self.island_expectation) > interval:
                partial_cumu = sum(self.island_expectation[-interval:-1])
            else:
                partial_cumu = sum(self.island_expectation)
            #for index in  xrange(len(self.island_expectation)):
            #print  index*self.bin_size, self.island_expectation[index];

        self.generate_cumulative_dist()
        for index in xrange(len(self.cumulative)):
            if self.cumulative[index] <= e_value_threshold:
                score_threshold = index * self.bin_size
                break
        return score_threshold

    def func(self, x):
        sum_ = 0.0
        for index in range(self.min_tags_in_window, self.max_index):
            sum_ += self.gap_contribution * pow(self.poisson_value[index],
                                                1 - x)
        return sum_ - 1

    def bracket_root(self, f, interval, max_iterations=50):
        """\
		Given a univariate function f and a tuple interval=(x1,x2),
		return a new tuple (bracket, fnvals) where bracket=(x1,x2)
		brackets a root of f and fnvals=(f(x1),f(x2)).
		"""
        GOLDEN = (1 + 5**.5) / 2
        (x1, x2) = interval
        if x1 == x2:
            raise Exception("initial interval has zero width")
        elif x2 < x1:
            x1, x2 = x2, x1
        f1, f2 = f(x1), f(x2)
        for j in xrange(max_iterations):
            while f1 * f2 >= 0:  # not currently bracketed
                if abs(f1) < abs(f2):
                    x1 = x1 + GOLDEN * (x1 - x2)
                else:
                    x2 = x2 + GOLDEN * (x2 - x1)
                f1, f2 = f(x1), f(x2)
            return (x1, x2), (f1, f2)
        raise Exception("too many iterations")

    # based on Numerical Recipes, p. 354
    def bisect_root(self, func, interval, xacc):
        JMAX = 50
        (x1, x2) = interval
        f = func(x1)
        fmid = func(x2)
        if (f * fmid >= 0.0): print("Root must be bracketed for bisection")
        if (f < 0.0):
            dx = x2 - x1
            rtb = x1
        else:
            dx = x1 - x2
            rtb = x2
        for j in xrange(JMAX):
            dx *= 0.5
            xmid = rtb + dx
            fmid = func(xmid)
            if (fmid <= 0.0): rtb = xmid
            if (fabs(dx) < xacc or fmid == 0.0): return rtb
        print("Too many bisections")
        return 0.0

    def find_asymptotics_exponent(self, xacc=.00001):
        num = 100
        #for index in xrange(num):
        #	x = index/float(num);
        #	print x, self.func(x);
        input_bracket = (0.1, 1)
        (xresult, yresult) = self.bracket_root(self.func, input_bracket)
        root = self.bisect_root(self.func, xresult, xacc)
        #print "# The exponent is: ", root;
        return root


# Species:  hg38
# Window_size:  200
# Gap size:  600
# E value is: 1000.0
# Total read count: 20227554.0
# Genome Length:  3088286401
# Effective genome Length:  2625043440
# Window average: 1.54112146807
# opt.evalue: 1000.0
# Total read count: 20227554.0
# window_size: 200
# opt.gap: 600
# Window pvalue: 0.2
# genome_length: 2625043440
# ('self.average=', 1.541121468069877)
# ('self.poisson_value[0]=', 0.2141408146291121)
# ('window_pvalue', 0.2)
# ('poisson', 0, 0.2141408146291121)
# ('poisson', 1, 0.33001700661489664)
# ('poisson', 2, 0.25429814686118796)
# ('poisson', 3, 0.13063477780605437)
# ('self.gap_contribution', 3.5943004243263443)
# ('self.boundary_contribution', 0.5552199656493835)
# ('island_expectation length', 2990)
# ('self.island_expectation[scaled_score]', 366781.3884759084)
# ('scaled_score', 2989)
# Minimum num of tags in a qualified window:  4
# Generate the enriched probscore summary graph and filter the summary graph to get rid of ineligible windows
# Determine the score threshold from random background
# opt.evalue 1000.0
# The score threshold is:  19.112
# Make and write islands
# Total number of islands:  74071

# Find candidate islands exhibiting clustering ...
# /mnt/work/endrebak/software/anaconda/envs/py27/bin/python /home/endrebak/code/epic_paper/SICER/src/find_islands_in_pr.py -s hg38 -b /mnt/scratch/projects/epic_bencmarks/data/sicer_results/0h_exp1/no_bigwig/0h_exp1_chip-W200.graph -w 200 -g 600 -t 0.85 -e 1000 -f /mnt/scra
# tch/projects/epic_bencmarks/data/sicer_results/0h_exp1/no_bigwig/0h_exp1_chip-W200-G600.scoreisland
# Species:  hg38
# Window_size:  200
# Gap size:  600
# E value is: 1000.0
# Total read count: 2792406.0
# Genome Length:  3088286401
# Effective genome Length:  2625043440
# Window average: 0.212751222128
# opt.evalue: 1000.0
# Total read count: 2792406.0
# window_size: 200
# opt.gap: 600
# Window pvalue: 0.2
# genome_length: 2625043440
# ('self.average=', 0.21275122212834696)
# ('self.poisson_value[0]=', 0.8083572135909076)
# ('window_pvalue', 0.2)
# ('poisson', 0, 0.8083572135909076)
# ('self.gap_contribution', 2.990012655388544)
# ('self.boundary_contribution', 0.18231673960386235)
# ('island_expectation length', 1761)
# ('self.island_expectation[scaled_score]', 411536.58828177786)
# ('scaled_score', 1760)
# Minimum num of tags in a qualified window:  1
# Generate the enriched probscore summary graph and filter the summary graph to get rid of ineligible windows
# ('read_count', 1.0, 'start', 1949600)
# ('read_count', 1.0, 'start', 1949800)
# ('read_count', 1.0, 'start', 1950000)
# ('read_count', 3.0, 'start', 1950200)
# ('read_count', 1.0, 'start', 1950600)
# ('read_count', 1.0, 'start', 1950800)
# ('read_count', 1.0, 'start', 1951000)
# ('read_count', 1.0, 'start', 1951600)
# ('read_count', 1.0, 'start', 1951800)
# ('read_count', 1.0, 'start', 1952600)
# ('read_count', 1.0, 'start', 1953400)
# ('read_count', 1.0, 'start', 1953600)
# ('read_count', 1.0, 'start', 1954000)
# Determine the score threshold from random background
# opt.evalue 1000.0
# The score threshold is:  27.363
# Make and write islands
# Total number of islands:  16431


def compute_score_threshold(chip_counts, window_size, effective_genome_length,
                            gap_size, e_value):

    tag_density = chip_counts / effective_genome_length
    background = Background_island_probscore_statistics(
        chip_counts, window_size, gap_size, 0.2, effective_genome_length,
        0.001)
    score_threshold = background.find_island_threshold(e_value)

    logging.info("Score threshold: {}\n".format(score_threshold))

    min_tags_in_window = background.min_tags_in_window

    average_window_readcount = chip_counts * (
        window_size / float(effective_genome_length))

    return score_threshold, min_tags_in_window, average_window_readcount



# require pysam is pre-installed
try:
    import pysam
except ImportError:
    raise Exception(
        '\n\n' +
        'pysam not found; please install pysam first (pip install pysam)' * 5 +
        '\n\n')

import os
import sys

from distutils.core import setup

from setuptools import find_packages, Extension, Command
from Cython.Build import cythonize

__version__ = open("epic2/version.py").readline().split(" = ")[1].replace(
    '"', '').strip()
macros = []

install_requires = [
    "scipy", "numpy", "natsort", "cython", "pysam", "pandas",
    "pyranges>=0.0.27"
]

if sys.version_info[0] == 2:
    install_requires.append("functools32")

if os.getenv("TRAVIS"):
    install_requires.append("coveralls")

if sys.version_info[0] == 2:
    install_requires.append("functools32")

compile_options = [
    "-Ofast", "-Wall", "-std=c++11"
]  #, "-frename-registers", "-funroll-loops"] # , "-lgzstream", "-lz"

from subprocess import check_output

try:
    conda_path = check_output("which conda", shell=True).decode().strip()
except:
    conda_path = ""

conda_include = []
conda_lib = []
if conda_path:
    conda_base = conda_path.replace("bin/conda", "")
    conda_include.append(os.path.join(conda_base, "include"))
    conda_lib.append(os.path.join(conda_base, "lib"))

dir_path = os.path.dirname(os.path.realpath(__file__))
include_dirs = [dir_path + "/epic2/src", dir_path]
# print(conda_lib)

extensions = [
    Extension(
        "epic2.src.reads_to_bins",
        ["epic2/src/reads_to_bins.pyx", "epic2/src/gzstream.cpp"],
        language="c++",
        include_dirs=conda_include + include_dirs,
        library_dirs=conda_lib,
        extra_compile_args=compile_options,
        libraries=["z"]),
    Extension(
        "epic2.src.SICER_stats", ["epic2/src/SICER_stats.pyx"],
        language="c++",
        extra_compile_args=compile_options),
    Extension(
        "epic2.src.SICER_stats2", ["epic2/src/SICER_stats2.pyx"],
        language="c++",
        extra_compile_args=compile_options),
    Extension(
        "epic2.src.statistics", ["epic2/src/statistics.pyx"],
        language="c++",
        extra_compile_args=compile_options),
    # Extension("epic2.src.differential",
    #           ["epic2/src/differential.pyx"], language="c++",
    #           extra_compile_args=compile_options),
    Extension(
        "epic2.src.find_islands", ["epic2/src/find_islands.pyx"],
        language="c++",
        extra_compile_args=compile_options),
    Extension(
        "epic2.src.read_bam", ["epic2/src/read_bam.pyx"],
        language="c++",
        extra_compile_args=compile_options,
        include_dirs=pysam.get_include() + conda_include,
        library_dirs=conda_lib,
        define_macros=pysam.get_defines(),
        libraries=["z"]),
    Extension(
        "epic2.src.genome_info", ["epic2/src/genome_info.pyx"],
        language="c++",
        extra_compile_args=compile_options)
]

setup(
    name="epic2",
    packages=find_packages(),
    ext_modules=cythonize(extensions, annotate=True, language_level='2'),
    scripts=["bin/epic2", "bin/epic2-df", "bin/epic2-bw"],
    package_data={
        'epic2': [
            'effective_sizes/*.txt', 'chromsizes/*chromsizes',
            'examples/*.bed.gz'
        ],
        '': ['*.pyx', '*.pxd', '*.h', '*.c', '*.hpp', "*.bed.gz"]
    },
    version=__version__,
    description="Ultraperformant ChIP-Seq broad peak/domain finder.",
    author="Endre Bakken Stovner",
    author_email="endrebak85@gmail.com",
    url="http://github.com/endrebak/epic2",
    keywords=["ChIP-Seq"],
    license=["MIT"],
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta", "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Topic :: Scientific/Engineering"
    ],
    include_dirs=["."],
    long_description=
    ("Chip-Seq broad peak/domain finder based on SICER. See the url for more info."
     ))



from __future__ import print_function

import logging
from collections import defaultdict
from cython.operator import dereference, postincrement



cimport epic2.src.cpp_read_files as cr

import sys
import numpy as np

from libc.stdint cimport uint32_t, uint16_t

from epic2.src.read_bam import read_bam
from epic2.src.genome_info import sniff

from cython.operator import dereference
from libcpp.algorithm cimport sort as stdsort
from libcpp.map cimport map as cppmap
from libcpp.algorithm cimport unique
from libcpp.vector cimport vector
from libcpp.string cimport string
from libcpp cimport bool
from libcpp.map cimport map

cdef extern from "<algorithm>" namespace "std" nogil:
    OutputIter merge[InputIter1, InputIter2, OutputIter] (InputIter1 first1, InputIter1 last1,
                                                          InputIter2 first2, InputIter2 last2,
                                                          OutputIter result)

cimport cython

import numpy as np

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cpdef remove_out_of_bounds_bins(count_dict, chromsizes, bin_size):

    # original code: get_bed_coords() in make_graph_file.py

    cdef:
        # uint32_t[::1] bins
        # uint16_t[::1] counts
        uint32_t chromsize
        size_t i


    for chromosome, (bins, counts) in count_dict.items():
        if not chromosome in chromsizes:
            print("Chromosome {} found in file, but not in the chromsizes dict for the genome.".format(chromosome), file=sys.stderr)
            continue

        chromsize = chromsizes[chromosome]

        # sys.stderr.write(chromosome + "\n")
        # sys.stderr.write("{} {}\n".format(bins[len(bins) - 1], chromsize))
        i = len(bins) - 1
        while (i >= 0 and (bins[i]) > chromsize):
            # print("For {} bin {} is out of bounds ({})\n".format(chromosome, bins[i], chromsize))
            # print("It has a count of ({})\n".format(counts[i]))
            # 52280 - 51862 = 418
            i -= 1

        if i != len(bins) - 1:
            # print("-----")
            # print(chromosome)
            # print("length before:", len(counts))
            count_dict[chromosome] = bins[:i+1], counts[:i+1]
            # print("length after:", len(counts[:i + 1]))



@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cpdef count_reads_per_bin(tags):

    cdef:
        uint32_t[::1] bins
        uint16_t[::1] counts
        Vector32 v
        uint32_t i
        uint32_t last = 1
        uint32_t current
        uint16_t count = 0
        uint32_t nparsed
        uint32_t last_two_equal

    bins_counts = dict()
    for k, v in tags.items():
        nparsed = 0
        count = 0

        bin_arr = np.ones(len(v), dtype=np.uint32)
        bins = bin_arr
        count_arr = np.ones(len(v), dtype=np.uint16)
        counts = count_arr

        if len(v) >= 1:
            last = v.wrapped_vector[0]

        for i in range(0, len(v)):
            current = v.wrapped_vector[i]

            if current != last:
                bins[nparsed] = last
                counts[nparsed] = count
                last = current
                count = 1
                nparsed += 1
            else:
                count += 1

        last_two_equal = v.wrapped_vector[len(v) - 2] == v.wrapped_vector[len(v) - 1]
        if last_two_equal:
            bins[nparsed] = v.wrapped_vector[len(v) - 1]
            counts[nparsed] = count
            nparsed += 1
        else:
            bins[nparsed] = v.wrapped_vector[len(v) - 1]
            counts[nparsed] = 1
            nparsed += 1


        #chrM    0       16399   52280   87916   1       0.419037470139  1
        # sys.stderr.write("Found {} in count_reads_per_bin\n".format(sum(count_arr[:nparsed])))
        bins_counts[k] = (bin_arr[:nparsed], count_arr[:nparsed])

    return bins_counts


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef class Vector32:

    cdef vector[uint32_t] wrapped_vector

    cdef push_back(self, uint32_t num):
        self.wrapped_vector.push_back(num)

    def sort(self):
        stdsort(self.wrapped_vector.begin(), self.wrapped_vector.end())

    def unique(self):
        self.wrapped_vector.erase(unique(self.wrapped_vector.begin(), self.wrapped_vector.end()), self.wrapped_vector.end())

    def __str__(self):
        return "[" + ", ".join([str(i) for i in self.wrapped_vector]) + "]"

    def __repr__(self):
        return str(self)

    def __len__(self):
        return self.wrapped_vector.size()

    def __iter__(self):
        # slow, only implemented to ease testing
        return (v for v in self.wrapped_vector)

    def merge(self, Vector32 other):

        cdef vector[uint32_t] o = vector[uint32_t](len(self) + len(other))
        merge(self.wrapped_vector.begin(), self.wrapped_vector.end(),
              other.wrapped_vector.begin(), other.wrapped_vector.end(),
              o.begin())

        cdef Vector32 output = Vector32()
        output.wrapped_vector = o

        return output




@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cpdef files_to_bin_counts(files, args, datatype):

    cdef:
        uint32_t bin_size = args["bin_size"]
        uint32_t half_fragment_size = args["fragment_size"] / 2
        uint32_t drop_duplicates = args["drop_duplicates"]
        Vector32 v
        Vector32 v2
        long[::1] bin_arr
        bytes py_bytes
        char* c_string
        cr.genome_map cpp_tags
        map[cr.key, cr.intvec].iterator it
        bool paired_end = False
        str file_format


    sum_tags = defaultdict(list)
    logging.info("Parsing {} file(s):".format(datatype))
    sys.stderr.flush()
    tags = dict()

    # uint32_t mapq, uint64_t required_flag, uint64_t filter_flag

    for f in files:

        paired_end = False

        logging.info("  " + f)
        sys.stderr.flush()

        py_bytes = f.encode()
        c_string = py_bytes

        file_format = sniff(f)

        if file_format == "bed":
            cpp_tags = cr.read_bed(c_string, drop_duplicates)
        elif file_format == "bedpe":
            cpp_tags = cr.read_bedpe(c_string, drop_duplicates)
            paired_end = True
        elif file_format == "bam": # sam also okay here
            cpp_tags = read_bam(f, drop_duplicates, args["mapq"], args["required_flag"], args["filter_flag"])
        elif file_format == "bed.gz":
            cpp_tags = cr.read_bed_gz(c_string, drop_duplicates)
        elif file_format == "bedpe.gz":
            cpp_tags = cr.read_bedpe_gz(c_string, drop_duplicates)
            paired_end = True

        it = cpp_tags.begin();

        while (it != cpp_tags.end()):
            chromosome = dereference(it).first.first.decode()

            if chromosome not in args["chromsizes_"].keys():
                postincrement(it)
                continue

            strand = chr(dereference(it).first.second)

            v = Vector32()
            v.wrapped_vector = dereference(it).second
            tags[chromosome, strand] = v

            postincrement(it)

        if not paired_end:
            for (chromosome, strand), v in tags.items():

                v.sort() # needs to be done again, since extracting the 5' end might make tags wrong order

                if strand == "+":
                    for i in range(len(v)):
                        v.wrapped_vector[i] = v.wrapped_vector[i] + half_fragment_size
                        v.wrapped_vector[i] = v.wrapped_vector[i] - (v.wrapped_vector[i] % bin_size)
                else:
                    i = 0
                    while i < len(v) and v.wrapped_vector[i] < half_fragment_size:
                        v.wrapped_vector[i] = 0
                        i += 1

                    for i in range(i, len(v)):
                        v.wrapped_vector[i] = v.wrapped_vector[i] - half_fragment_size
                        v.wrapped_vector[i] = v.wrapped_vector[i] - (v.wrapped_vector[i] % bin_size)

                # for i in range(len(v)):
                #     v.wrapped_vector[i] = v.wrapped_vector[i] - (v.wrapped_vector[i] % bin_size)

                # sys.stderr.write("Found {} for {} {}\n".format(i, chromosome, strand))
                if chromosome not in sum_tags:
                    sum_tags[chromosome] = v
                else:
                    v2 = sum_tags[chromosome]
                    sum_tags[chromosome] = v.merge(v2)

        else:

            for (chromosome, strand), v in tags.items():

                v.sort() # needs to be done again, since extracting the 5' end might make tags wrong order

                for i in range(len(v)):
                    v.wrapped_vector[i] = v.wrapped_vector[i] - (v.wrapped_vector[i] % bin_size)

                if chromosome not in sum_tags:
                    sum_tags[chromosome] = v
                else:
                    v2 = sum_tags[chromosome]
                    sum_tags[chromosome] = v.merge(v2)


    sys.stderr.flush()

    bins_counts = count_reads_per_bin(sum_tags)

    # import pandas as pd
    # bins, counts = [pd.Series(s) for s in bins_counts["chrY"]]
    # print(bins.tail(), file=sys.stderr)
    # print(counts.tail(), file=sys.stderr)

    count = sum([sum(counts) for _, counts in bins_counts.values()])
    remove_out_of_bounds_bins(bins_counts, args["chromsizes_"], bin_size)

    # bins, counts = [pd.Series(s) for s in bins_counts["chrY"]]
    # print(bins.tail())
    # print(counts.tail())

    # print(bins.tail(), file=sys.stderr)
    # print(counts.tail(), file=sys.stderr)

    return bins_counts, count



cpdef add_reads_to_dict(f, chromosomes):

    genome = dict()
    cdef Vector32 v

    for line in open(f):
        chromosome, left, right, _, _, strand = line.split()

        if chromosome not in chromosomes:
            continue

        if strand == "+":
            five_end = <uint32_t> int(left)
        else:
            five_end = <uint32_t> int(right)

        if (chromosome, strand) in genome:
            v = genome[chromosome, strand]
            v.wrapped_vector.push_back(five_end)
        else:
            v = Vector32()
            v.wrapped_vector.push_back(five_end)
            genome[chromosome, strand] = v

    return genome



######## this was actually slower!

# cpdef add_reads_to_dict(f):

#     genome = dict()
#     cdef:
#         Vector v
#         FILE *f_handle
#         char chromosome [10]
#         char strand [2]
#         uint32_t left
#         uint32_t right

#     fp = fopen(f.encode(), "r")

#     while (
#             fscanf(fp, "%s\t%d\t%d\t%*s\t%*d\t%s\n", chromosome, &left, &right, strand) != EOF
#     ):


#         # pruint32_t("----")
#         # pruint32_t("chromosome is", chromosome)
#         # pruint32_t("strand is", strand)
#         # pruint32_t("left is", left)
#         # pruint32_t("right is", right)
#         if strand == b"+":
#             five_end = left
#         else:
#             five_end = right

#         # pruint32_t("five end is ", five_end)

#         if (chromosome, strand) in genome:
#             v = genome[chromosome, strand]
#             v.wrapped_vector.push_back(five_end)
#         else:
#             v = Vector()
#             v.wrapped_vector.push_back(five_end)
#             genome[chromosome, strand] = v

#     return genome



from __future__ import print_function

from scipy.stats import poisson, rankdata
import sys
import logging
import cython
from libcpp cimport bool

from libcpp.algorithm cimport sort as stdsort
from libcpp.string cimport string
from libc.math cimport log2
import numpy as np

from libc.stdint cimport uint32_t, uint16_t, int32_t
from epic2.src.statistics import compute_window_score
from libcpp.vector cimport vector
# from libc.math cimport min as cmin

from natsort import natsorted


cdef struct island:
    string chromosome
    uint32_t start
    uint32_t end
    uint32_t chip_count
    uint32_t input_count
    double score
    double p_value
    double fold_change


# cdef bool compare_p_value(const island &lhs, const island &rhs):

#     return lhs.p_value < rhs.p_value



cdef class IslandVector():

    cdef vector[island] wrapped_vector

    cdef push_back(self, island i):
        self.wrapped_vector.push_back(i)

    def __str__(self):
        return "[" + ", ".join([str(i) for i in self.wrapped_vector]) + "]"

    def __repr__(self):
        return str(self)

    def __len__(self):

        return int(self.wrapped_vector.size())

    def __iter__(self):
        # slow, only implemented to ease testing
        return (v for v in self.wrapped_vector)

    # def p_value_sort(self):
    #     stdsort(self.wrapped_vector.begin(), self.wrapped_vector.end(), &compare_p_value)



@cython.boundscheck(False)
@cython.wraparound(True)
@cython.initializedcheck(False)
def find_islands(bins_counts, int gaps_allowed, int bin_size, float score_threshold, uint32_t min_tags_in_window, float average_window_readcount):

    _poisson = poisson(average_window_readcount)

    cdef:
        double slightly_less = score_threshold - 0.0000000001
        double score
        int i
        int length
        uint32_t count
        uint32_t _bin
        uint32_t[::1] bins = np.ones(1, dtype=np.uint32)
        uint16_t[::1] counts = np.ones(1, dtype=np.uint16)
        int32_t dist # distance can be negative on the very first island in each chromosome, hence not uint
        island current_island
        IslandVector v
        int32_t distance_allowed = (gaps_allowed * bin_size) + 2

    chromosomes = natsorted(set(bins_counts.keys()))

    island_dict = dict()

    for chromosome in chromosomes:
        v = IslandVector()
        # print("-------")
        # print(chromosome)
        i = 0

        bins, counts = bins_counts[chromosome]

        length = len(bins)

        # print("length", length)
        while i < length and (counts[i] < min_tags_in_window):
            i += 1

        # print("first i, counts", i, counts[i])

        count = 0
        _bin = bins[i]
        score = compute_window_score(count, _poisson)

        chromosome = chromosome.encode("utf-8")
        current_island = [chromosome, _bin, _bin + bin_size - 1, 0, 0, score, 0, 0]
        while i < length:

            if (counts[i] < min_tags_in_window):
                i += 1
                continue

            _bin = bins[i]


            dist = _bin - current_island.end

            # # _print = chromosome == b"chrM"# and _bin >= 107040800
            # if _print:
            #     print("_bin", _bin, "i", i)
            #     print("_count", counts[i])
            #     print("dist", dist)
            #     print("current_island", current_island)
            #     print("dist allowed", distance_allowed)

            if dist <= distance_allowed:

                # if _print:
                #     print("we are in dist")

                current_island.end = bins[i] + bin_size - 1
                # current_island.chip_count += counts[i]
                current_island.score += compute_window_score(counts[i], _poisson)
            else:
                if current_island.score > slightly_less: # and current_island.chip_count > min_tags_in_windw:
                    # if _print:
                    #     print("pushed back", current_island)
                    v.push_back(current_island)

                # here need to iterate until we get to a
                # bin with at least min_tags_in_window number of reads
                while i < length and (counts[i] < min_tags_in_window):
                    i += 1
                # print("second i, counts", i, counts[i])

                # if _print:
                #     print("current_island_before", current_island)

                score = compute_window_score(counts[i], _poisson)
                current_island = [chromosome, _bin, bins[i] + bin_size - 1, 0, 0, score, 0, 0]

                # if _print:
                #     print("current_island_after", current_island)
                # else: # then length == i


            i += 1

        # takes care of last or potential single bin
        if current_island.score > slightly_less: # and current_island.chip_count > min_tags_in_windw:
            # print("current_island last", current_island)
            v.push_back(current_island)

        island_dict[chromosome.decode()] = v
        # print(chromosome, v.wrapped_vector[0])

    return island_dict


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
@cython.cdivision(False)
def add_chip_count_to_islands(islands, c_bins_counts, args):

    cdef:
        int i
        int j = 0
        island current_island
        IslandVector _islands
        IslandVector updated_islands
        uint32_t[::1] bins = np.ones(1, dtype=np.uint32)
        uint16_t[::1] counts = np.ones(1, dtype=np.uint16)


    new_islands = {}
    chromosomes = natsorted(set(islands.keys()))

    for chromosome in chromosomes:

        updated_islands = IslandVector()
        j = 0
        _islands = islands[chromosome]
        # print("chr", chromosome, len(_islands))
        # print(chromosome, c_bins_counts)

        if chromosome not in c_bins_counts:
            # print("chromosome", chromosome, "not in", c_bins_counts)
            continue

        bins, counts = c_bins_counts[chromosome]
        number_bins = len(bins)

        for i in range(len(_islands)):
            # _island = &_islands.wrapped_vector[i]
            _island = _islands.wrapped_vector[i]

            # not overlapping
            while j < number_bins and (bins[j] < _island.start):
                # print("not overlapping:")
                # print("bins[j]", bins[j], "counts[j]", counts[j], "_island.start", _island)
                j += 1

            # overlapping
            while j < number_bins and (bins[j] < _island.end and bins[j] >= _island.start):
                # print("overlapping:")
                # print("bins[j]", bins[j], "counts[j]", counts[j], "_island.start", _island)
                _island.chip_count += counts[j]
                j += 1

            updated_islands.wrapped_vector.push_back(_island)

        new_islands[chromosome] = updated_islands

        if not args.get("df"):
            del c_bins_counts[chromosome]

    return new_islands

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
@cython.cdivision(True)
def compute_fdr(islands, b_bins_counts, int chip_library_size, int control_library_size, double effective_genome_fraction, double fdr_cutoff, output):

    cdef:
        int i
        int j
        int number_bins
        island _island
        IslandVector _islands
        IslandVector all_islands = IslandVector()
        uint16_t count
        uint32_t _bin
        uint32_t[::1] bins = np.ones(1, dtype=np.uint32)
        uint16_t[::1] counts = np.ones(1, dtype=np.uint16)
        double scaling_factor = float(chip_library_size) / float(control_library_size)
        double zero_scaler = (control_library_size / effective_genome_fraction)
        double fdr
        double average
        int num_islands
        int counter

    sf = poisson.sf
    chromosomes = natsorted(set(islands.keys()))

    num_islands_per_chrom = {}
    for c in natsorted(islands):
        _islands = islands[c]
        num_islands_per_chrom[c] = len(_islands)

    for chromosome in chromosomes:

        j = 0
        _islands = islands[chromosome]
        # print("chr", chromosome, len(_islands), _islands)

        if chromosome not in b_bins_counts:
            continue

        bins, counts = b_bins_counts[chromosome]
        number_bins = len(bins)

        for i in range(len(_islands)):

            # not overlapping
            while j < number_bins and (bins[j] < _islands.wrapped_vector[i].start):
                # print("bins[j]", bins[j], "_islands.wrapped_vector[i].start", _islands.wrapped_vector[i].start)
                j += 1

            # overlapping
            while j < number_bins and (bins[j] < _islands.wrapped_vector[i].end and bins[j] >= _islands.wrapped_vector[i].start):
                _islands.wrapped_vector[i].input_count += counts[j]
                j += 1

            # if _islands.wrapped_vector[i].start == 117299800:
            #     _print = True

            # else:
            #     _print = False

            ##### COMPUTE P ########
            if _islands.wrapped_vector[i].input_count > 0:
                average = _islands.wrapped_vector[i].input_count * scaling_factor
                # if _print:
                #     print("input_count", _islands.wrapped_vector[i].input_count)
                #     print("scaling factor", scaling_factor)
                #     print("average", average)
            else:
                average = (_islands.wrapped_vector[i].end - _islands.wrapped_vector[i].start + 1) * zero_scaler
                average = min(0.25, average) * scaling_factor

            _islands.wrapped_vector[i].fold_change = log2(_islands.wrapped_vector[i].chip_count / average)

            if _islands.wrapped_vector[i].chip_count > average:
                _islands.wrapped_vector[i].p_value = sf(_islands.wrapped_vector[i].chip_count, average)
                # if _print:
                #     print("p_val:", sf(_islands.wrapped_vector[i].chip_count, average))
                #     print("chip_count:", _islands.wrapped_vector[i].chip_count)
            else:
                _islands.wrapped_vector[i].p_value = 1


            all_islands.push_back(_islands.wrapped_vector[i])

        del b_bins_counts[chromosome]

    p_values = np.zeros(len(all_islands))

    for i in range(len(all_islands)):
        p_values[i] = all_islands.wrapped_vector[i].p_value

    ranks = rankdata(p_values)

    counter = 0
    j = 0
    num_islands = len(all_islands)

    if not output:
        handle = sys.stdout
    else:
        handle = open(output, "w+")

    print("\t".join(["#Chromosome", "Start", "End", "PValue", "Score", "Strand", "ChIPCount", "InputCount", "FDR", "log2FoldChange"]), file=handle)
    for i in range(num_islands):
            # _island = all_islands.wrapped_vector[i]
            fdr = all_islands.wrapped_vector[i].p_value * num_islands / ranks[i]
            if fdr > 1:
                fdr = 1

            if fdr <= fdr_cutoff:
                chromosome = all_islands.wrapped_vector[i].chromosome.decode()
                print("\t".join(str(e) for e in [chromosome, all_islands.wrapped_vector[i].start, all_islands.wrapped_vector[i].end, all_islands.wrapped_vector[i].p_value,
                                                 min(1000, all_islands.wrapped_vector[i].fold_change * 100), ".", all_islands.wrapped_vector[i].chip_count, all_islands.wrapped_vector[i].input_count, fdr, all_islands.wrapped_vector[i].fold_change]), file=handle)

    if output:
        handle.close()


def write_islands(islands, float average_window_readcount, float fdr_cutoff, int e_value, output):

    cdef:
        IslandVector _islands
        IslandVector all_islands = IslandVector()
        int i
        int num_islands = 0
        int counter


    chromosomes = natsorted(set(islands.keys()))

    num_islands_per_chrom = {}
    for c in natsorted(islands):
        _islands = islands[c]
        num_islands_per_chrom[c] = len(_islands)


    _poisson = poisson(average_window_readcount)

    if not output:
        handle = sys.stdout
    else:
        handle = open(output, "w+")

    print("\t".join(["#Chromosome", "Start", "End", "ChIPCount", "Score", "Strand"]), file=handle)
    for chromosome in chromosomes:
        _islands = islands[chromosome]
        num_islands += len(_islands)

        for i in range(len(_islands)):
            print("\t".join(str(e) for e in [chromosome, _islands.wrapped_vector[i].start, _islands.wrapped_vector[i].end, _islands.wrapped_vector[i].chip_count, float(_islands.wrapped_vector[i].score), "."]),
                  file=handle)

    if output:
        handle.close()

    logging.info("Empirical estimate of FDR is: {} ({}/{})".format(min(1, float(e_value)/num_islands), e_value, num_islands))
    # ranks = rankdata(np.array([_islands.wrapped_vector[i].p_value for _islands.wrapped_vector[i] in all_islands.wrapped_vector[i]s], dtype=np.int))

    # p_values = np.zeros(len(all_islands))

    # for i in range(len(all_islands)):
    #     p_values[i] = all_islands.wrapped_vector[i].p_value

    # ranks = rankdata(p_values)

    # counter = 0
    # num_islands = len(all_islands)

    # for chromosome, chromosome_size in natsorted(num_islands_per_chrom.items()):

    #     i = 0
    #     for i in range(int(chromosome_size)):

    #         #### compute fdr #####
    #         # _islands.wrapped_vector[i] = all_islands.wrapped_vector[i + counter]

    #         fdr = all_islands.wrapped_vector[i + counter].p_value * num_islands / ranks[i + counter]
    #         # print("fdr cutoff", fdr_cutoff)
    #         if fdr > 1:
    #             fdr = 1
    #         if fdr <= fdr_cutoff:
    #             print("\t".join(str(e) for e in [chromosome, all_islands.wrapped_vector[i].start, all_islands.wrapped_vector[i].end, all_islands.wrapped_vector[i].p_value, all_islands.wrapped_vector[i].chip_count, ".", fdr]))


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
@cython.cdivision(False)
def differential_count_reads_on_islands(dfs, c_bins_counts):

    # want count per island
    # must iterate over islands, per chromosome
    # add counts per islands

    # need vector to hold counts

    cdef:
        int i, number_bins, number_islands
        int j = 0
        uint32_t[::1] bins = np.ones(1, dtype=np.uint32)
        uint16_t[::1] counts = np.ones(1, dtype=np.uint16)
        uint32_t[::1] starts
        uint32_t[::1] ends

    new_islands = {}
    chromosomes = natsorted(dfs)

    for chromosome in chromosomes:
        df = dfs[chromosome]

        j = 0
        i = 0

        number_islands = len(df)
        outcounts = np.zeros(number_islands, dtype=np.uint32)
        starts, ends = df.Start.astype(np.uint32).values, df.End.astype(np.uint32).values

        if chromosome not in c_bins_counts:
            new_islands[chromosome] = outcounts
            continue

        bins, counts = c_bins_counts[chromosome]
        number_bins = len(bins)

        for i in range(number_islands):
            # _island = &_islands.wrapped_vector[i]

            # not overlapping
            while j < number_bins and (bins[j] < starts[i]):
                # print("not overlapping:")
                # print("bins[j]", bins[j], "counts[j]", counts[j], "_island.start", _island)
                j += 1

            # overlapping
            while j < number_bins and (bins[j] < ends[i] and bins[j] >= starts[i]):
                # print("overlapping:")
                # print("bins[j]", bins[j], "counts[j]", counts[j], "_island.start", _island)
                outcounts[i] += counts[j]
                j += 1


        new_islands[chromosome] = outcounts

        # if not args.get("df"):
        #     del c_bins_counts[chromosome]

    return new_islands
